import time

from woger import BaseWorkspace


def _load_xml(fs, ws):
    fs.CreateFile(ws.raw_path)


def _parse_xml(fs, ws):
    fs.CreateFile(ws.parsed_path)


def test_manager_workspace_update(fs):
    from woger import WorkspaceManager, BaseAction

    class Workspace(BaseWorkspace):
        raw_path = 'data.xml'
        parsed_path = 'data.json'

    class Action(BaseAction):
        load_xml = 'load_data'
        parse_xml = 'parse_xml'

    manager = WorkspaceManager('root', Workspace)
    assert manager.current() is None

    latest_finished_ws = manager.find_latest_finished(Action.parse_xml)
    assert latest_finished_ws is None

    timestamp = 1346134
    manager.create(timestamp)
    manager.target_latest()

    ws = manager.target()
    tracker = ws.track(Action.load_xml)

    if not tracker.finished():
        assert tracker.started() or tracker.failed() or tracker.undefined()
        with tracker:
            assert tracker.started()
            _load_xml(fs, ws)
        assert tracker.finished()

    tracker = ws.track(Action.parse_xml)

    if not tracker.finished():
        with tracker:
            assert tracker.started()
            _parse_xml(fs, ws)
        assert tracker.finished()

    assert manager.current() is None
    manager.update()
    assert manager.current() == manager.storage[timestamp] == manager.target()



from .base_workspace import BaseWorkspace
from .state import WorkspaceState
from .storage import WorkspaceStorage
from .manager import WorkspaceManager
from .base_action import BaseAction
from .workspace_meta import WorkspaceMeta
from .action_tracker import ActionTracker
from .action_status import ActionStatus


__version__ = '1.0.6'

__all__ = [
    'BaseWorkspace',
    'WorkspaceState',
    'WorkspaceStorage',
    'WorkspaceManager',
    'BaseAction',
    'WorkspaceMeta',
    'ActionTracker',
    'ActionStatus',
]



from .action_status import ActionStatus
from .base_action import BaseAction
# from .base_workspace import BaseWorkspace


class ActionTracker:
    """Action management class"""
    def __init__(self, action: BaseAction, workspace):
        """Initialize action tracker

        Parameters
        ----------
        action: subclass of BaseAction
            Action to track
        workspace: subclass of Workspace
            Workspace to track action in
        """
        self.action = action
        self.workspace = workspace

    def __enter__(self):
        self.workspace.state[self.action.value] = ActionStatus.started.value

    def __exit__(self, exc_type, exc_val, exc_tb):
        status = (
            ActionStatus.finished
            if exc_tb is None
            else ActionStatus.failed
        )
        self.workspace.state[self.action.value] = status.value

    def _status_flag(self, status):
        return self.workspace.state[self.action.value] == status.value

    def started(self):
        """Returns True if action has started and is pending"""
        return self._status_flag(ActionStatus.started)

    def finished(self):
        """Returns True if action has finished"""
        return self._status_flag(ActionStatus.finished)

    def failed(self):
        """Returns True if action has failed"""
        return self._status_flag(ActionStatus.failed)

    def undefined(self):
        return self.workspace.state[self.action.value] is None



from pathlib import Path


def test_manager_latest_workspace(fs):
    from woger import WorkspaceManager, BaseWorkspace

    class Workspace(BaseWorkspace):
        def __lt__(self, other):
            return self.state_path < other.state_path

    root = Path('/') / 'global'
    subdirs = [
        root / name
        for name in ['134', '235', '9', '17']
    ]
    for subdir in subdirs:
        fs.CreateDirectory(subdir)

    wm = WorkspaceManager(root, Workspace)

    assert wm.latest().id == 235


def test_manager_current_and_target(fs):
    from woger import WorkspaceManager, BaseWorkspace

    root = Path('/') / 'global'
    ws1_path = root / str(346134)
    ws2_path = root / str(824842)
    state_path = root / '.state.json'

    fs.CreateFile(
        state_path,
        contents='\n'.join([
            '{',
            '"current": 346134,',
            '"target": 824842',
            '}'
        ])
    )
    fs.CreateDirectory(ws1_path)
    fs.CreateDirectory(ws2_path)

    class Workspace(BaseWorkspace):
        pass

    manager = WorkspaceManager(
        root,
        Workspace,
    )

    assert manager.state_path == state_path
    assert manager.current().path == ws1_path
    assert manager.target().path == ws2_path


def test_manager_search_latest_finished(fs):
    from woger import WorkspaceManager, BaseWorkspace, BaseAction

    root = Path('root')

    fs.CreateFile(
        root / '346' / '.state.json',
        contents='{"load_raw_xml":"finished","parse_xml": "finished"}',
    )
    fs.CreateFile(
        root / '134' / '.state.json',
        contents='{"load_raw_xml":"finished","parse_xml": "finished"}',
    )
    fs.CreateFile(
        root / '576' / '.state.json',
        contents='{"load_raw_xml": "finished","parse_xml": "started"}',
    )

    class Action(BaseAction):
        load_raw_xml = 'load_raw_xml'
        parse_xml = 'parse_xml'

    wm = WorkspaceManager(root, BaseWorkspace)
    ws = wm.find_latest_finished(Action.parse_xml)

    assert ws.id == 346


def test_manager_search_latest_finished_none(fs):
    from woger import WorkspaceManager, BaseWorkspace, BaseAction

    root = Path('root')

    fs.CreateFile(
        root / '346' / '.state.json',
        contents='{"load_raw_xml":"finished","parse_xml": "failed"}',
    )
    fs.CreateFile(
        root / '134' / '.state.json',
        contents='{"load_raw_xml": "started"}',
    )
    fs.CreateFile(
        root / '576' / '.state.json',
        contents='{"load_raw_xml": "finished","parse_xml": "started"}',
    )

    class Action(BaseAction):
        load_raw_xml = 'load_raw_xml'
        parse_xml = 'parse_xml'

    wm = WorkspaceManager(root, BaseWorkspace)
    ws = wm.find_latest_finished(Action.parse_xml)

    assert ws is None



import warnings
from pathlib import Path

import os
from typing import Optional,  Type

from .storage import WorkspaceStorage
from .state import WorkspaceState
from .base_workspace import BaseWorkspace

_TARGET = 'target'
_CURRENT = 'current'

def _load_workspaces(root_path: Path, workspace_cls):
    path_list = [
        root_path / path
        for path in os.listdir(root_path)
        if os.path.isdir(root_path / path)
    ]
    return map(workspace_cls, path_list)


class WorkspaceManager:
    """Manages workspaces

    Allows you to
    - detect workspaces
    - manage current, target and latest workspaces
    - track workspace actions
    - search for workspaces with finished actions
    """
    def __init__(self, path, workspace_cls: Type[BaseWorkspace]):
        """Initialize workspace manager

        Parameters
        ----------
        path: path-like
            Workspace storage path
        workspace_cls: class object of BaseWorkspace subclass
            Workspace class object
        """
        self.path = Path(path)
        self._state_path = Path('.state.json')
        os.makedirs(self.path, exist_ok=True)

        self.workspace_cls = workspace_cls

        workspaces = _load_workspaces(self.path, workspace_cls)
        self.storage = WorkspaceStorage(workspaces)

        self.state = WorkspaceState(self.state_path)

    @property
    def current_ws_id(self):
        return self.state[_CURRENT]

    @current_ws_id.setter
    def current_ws_id(self, value):
        import traceback
        traceback.print_stack()
        self.state[_CURRENT] = value

    @property
    def target_ws_id(self):
        return self.state[_TARGET]

    @target_ws_id.setter
    def target_ws_id(self, value):
        self.state[_TARGET] = value

    @property
    def state_path(self):
        """Path to the state file

        Is used internally
        """
        return self.path / self._state_path

    def create(self, name) -> BaseWorkspace:
        """Creates a Workspace

        Parameters
        ----------
        name: str or int
            Workspace name
        """
        if isinstance(name, int):
            name = str(name)
        path = self.path / name
        ws = self.workspace_cls(path)
        self.storage.add(ws)
        return ws

    def latest(self) -> BaseWorkspace:
        """Returns latest workspace"""
        return self.storage.at(-1)

    def current(self) -> Optional[BaseWorkspace]:
        """Returns current workspace"""
        return self.storage[self.current_ws_id]

    def target(self) -> BaseWorkspace:
        """Returns target workspace"""
        return self.storage[self.target_ws_id]

    def __repr__(self):
        return '<{self.__class__.__name__}(path={self.path}, storage={self.storage})>'.format(self=self)

    def find_latest_finished(self, action) -> Optional[BaseWorkspace]:
        """Searches for latest workspace with finished action constraint"""
        workspaces = list(self.storage.values())
        for ws in reversed(workspaces):
            tracker = ws.track(action)
            if tracker.finished():
                return ws

    def target_latest(self):
        self.target_ws_id = self.latest().id

    def update(self):
        """Move from current workspace to target one"""
        if self.current_ws_id == self.target_ws_id:
            warnings.warn('Current and target workspaces are the same')
            return

        self.current_ws_id = self.target_ws_id
